@@ -31,23 +31,29 @@
 #include <QPixmapCache>
 #include <QFile>
 #include <QMessageBox>
+#include <QCommandLineParser>
+#include <QSocketNotifier>
 #include <gio/gio.h>
+#include <sys/socket.h>
 
 #include "applicationadaptor.h"
 #include "preferencesdialog.h"
 #include "desktoppreferencesdialog.h"
-#include "mountoperation.h"
+#include <libfm-qt/mountoperation.h>
 #include "autorundialog.h"
 #include "launcher.h"
+#include <libfm-qt/filesearchdialog.h>
 
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
 #include <QScreen>
 #include <QWindow>
-#endif
 
 #include <X11/Xlib.h>
 
-using namespace PCManFM;
+#include "xdgdir.h"
+#include <QFileSystemWatcher>
+
+namespace PCManFM {
+
 static const char* serviceName = "org.pcmanfm.PCManFM";
 static const char* ifaceName = "org.pcmanfm.Application";
 
@@ -68,11 +74,15 @@ Application::Application(int& argc, char** argv):
   enableDesktopManager_(false),
   preferencesDialog_(),
   volumeMonitor_(NULL),
+  userDirsWatcher_(NULL),
+  lxqtRunning_(false),
   editBookmarksialog_() {
 
   argc_ = argc;
   argv_ = argv;
 
+  setApplicationVersion(PCMANFM_QT_VERSION);
+
   // QDBusConnection::sessionBus().registerObject("/org/pcmanfm/Application", this);
   QDBusConnection dbus = QDBusConnection::sessionBus();
   if(dbus.registerService(serviceName)) {
@@ -84,7 +94,9 @@ Application::Application(int& argc, char** argv):
     new ApplicationAdaptor(this);
     dbus.registerObject("/Application", this);
 
-    connect(this, SIGNAL(aboutToQuit()), SLOT(onAboutToQuit()));
+    connect(this, &Application::aboutToQuit, this, &Application::onAboutToQuit);
+    // aboutToQuit() is not signalled on SIGTERM, install signal handler
+    installSigtermHandler();
     settings_.load(profileName_);
 
     // decrease the cache size to reduce memory usage
@@ -94,6 +106,21 @@ Application::Application(int& argc, char** argv):
       QIcon::setThemeName(settings_.fallbackIconThemeName());
       Fm::IconTheme::checkChanged();
     }
+
+    // Check if LXQt Session is running. LXQt has it's own Desktop Folder
+    // editor. We just hide our editor when LXQt is running.
+    QDBusInterface* lxqtSessionIface = new QDBusInterface(
+                                        QStringLiteral("org.lxqt.session"),
+                                        QStringLiteral("/LXQtSession"));
+    if (lxqtSessionIface) {
+      if (lxqtSessionIface->isValid()) {
+          lxqtRunning_ = true;
+          userDesktopFolder_ = XdgDir::readDesktopDir();
+          initWatch();
+      }
+      delete lxqtSessionIface;
+      lxqtSessionIface = 0;
+    }
   }
   else {
     // an service of the same name is already registered.
@@ -110,140 +137,113 @@ Application::~Application() {
     g_object_unref(volumeMonitor_);
   }
 
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
-  if(enableDesktopManager_)
-    removeNativeEventFilter(this);
-#endif
+  // if(enableDesktopManager_)
+  //   removeNativeEventFilter(this);
 }
 
-struct FakeTr {
-  FakeTr(int reserved = 20) {
-    strings.reserve(reserved);
+void Application::initWatch()
+{
+  QFile file_ (QStandardPaths::writableLocation(QStandardPaths::ConfigLocation) + QStringLiteral("/user-dirs.dirs"));
+  if(! file_.open(QIODevice::ReadOnly | QIODevice::Text)) {
+    qDebug() << Q_FUNC_INFO << "Could not read: " << userDirsFile_;
+    userDirsFile_ = QString();
+  } else {
+    userDirsFile_ = file_.fileName();
   }
 
-  const char* operator()(const char* str) {
-    QString translated = QApplication::translate(NULL, str);
-    strings.push_back(translated.toUtf8());
-    return strings.back().constData();
-  }
-  QVector<QByteArray> strings;
-};
+  userDirsWatcher_ = new QFileSystemWatcher(this);
+  userDirsWatcher_->addPath(userDirsFile_);
+  connect(userDirsWatcher_, &QFileSystemWatcher::fileChanged, this, &Application::onUserDirsChanged);
+}
 
 bool Application::parseCommandLineArgs() {
   bool keepRunning = false;
+  QCommandLineParser parser;
+  parser.addHelpOption();
+  parser.addVersionOption();
 
-  // It's really a shame that the great Qt library does not come
-  // with any command line parser.
-  // After trying some Qt ways, I finally realized that glib is the best.
-  // Simple, efficient, effective, and does not use signal/slot!
-  // The only drawback is the translated string returned by tr() is
-  // a temporary one. We need to store them in a list to keep them alive. :-(
+  QCommandLineOption profileOption(QStringList() << "p" << "profile", tr("Name of configuration profile"), tr("PROFILE"));
+  parser.addOption(profileOption);
 
-  char* profile = NULL;
-  gboolean daemon_mode = FALSE;
-  gboolean ask_quit = FALSE;
-  gboolean desktop = FALSE;
-  gboolean desktop_off = FALSE;
-  char* desktop_pref = NULL;
-  char* wallpaper = NULL;
-  char* wallpaper_mode = NULL;
-  char* show_pref = NULL;
-  gboolean new_window = FALSE;
-  gboolean find_files = FALSE;
-  char** file_names = NULL;
-  {
-    FakeTr tr; // a functor used to override QObject::tr().
-    // it convert the translated strings to UTF8 and add them to a list to
-    // keep them alive during the option parsing process.
-    GOptionEntry option_entries[] = {
-      /* options only acceptable by first pcmanfm instance. These options are not passed through IPC */
-      {"profile", 'p', 0, G_OPTION_ARG_STRING, &profile, tr("Name of configuration profile"), tr("PROFILE") },
-      {"daemon-mode", 'd', 0, G_OPTION_ARG_NONE, &daemon_mode, tr("Run PCManFM as a daemon"), NULL },
-      // options that are acceptable for every instance of pcmanfm and will be passed through IPC.
-      {"quit", 'p', 0, G_OPTION_ARG_NONE, &ask_quit, tr("Quit PCManFM"), NULL},
-      {"desktop", '\0', 0, G_OPTION_ARG_NONE, &desktop, tr("Launch desktop manager"), NULL },
-      {"desktop-off", '\0', 0, G_OPTION_ARG_NONE, &desktop_off, tr("Turn off desktop manager if it's running"), NULL },
-      {"desktop-pref", '\0', 0, G_OPTION_ARG_STRING, &desktop_pref, tr("Open desktop preference dialog on the page with the specified name"), tr("NAME") },
-      {"set-wallpaper", 'w', 0, G_OPTION_ARG_FILENAME, &wallpaper, tr("Set desktop wallpaper from image FILE"), tr("FILE") },
-      // don't translate list of modes in description, please
-      {"wallpaper-mode", '\0', 0, G_OPTION_ARG_STRING, &wallpaper_mode, tr("Set mode of desktop wallpaper. MODE=(color|stretch|fit|center|tile)"), tr("MODE") },
-      {"show-pref", '\0', 0, G_OPTION_ARG_STRING, &show_pref, tr("Open Preferences dialog on the page with the specified name"), tr("NAME") },
-      {"new-window", 'n', 0, G_OPTION_ARG_NONE, &new_window, tr("Open new window"), NULL },
-      {"find-files", 'f', 0, G_OPTION_ARG_NONE, &find_files, tr("Open Find Files utility"), NULL },
-      {G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_FILENAME_ARRAY, &file_names, NULL, tr("[FILE1, FILE2,...]")},
-      { NULL }
-    };
+  QCommandLineOption daemonOption(QStringList() << "d" << "daemon-mode", tr("Run PCManFM as a daemon"));
+  parser.addOption(daemonOption);
 
-    GOptionContext* context = g_option_context_new("");
-    g_option_context_add_main_entries(context, option_entries, NULL);
-    GError* error = NULL;
-    if(!g_option_context_parse(context, &argc_, &argv_, &error)) {
-      // show error and exit
-      g_fprintf(stderr, "%s\n\n", error->message);
-      g_error_free(error);
-      g_option_context_free(context);
-      return false;
-    }
-    g_option_context_free(context);
-  }
+  QCommandLineOption quitOption(QStringList() << "q" << "quit", tr("Quit PCManFM"));
+  parser.addOption(quitOption);
+
+  QCommandLineOption desktopOption("desktop", tr("Launch desktop manager"));
+  parser.addOption(desktopOption);
+
+  QCommandLineOption desktopOffOption("desktop-off", tr("Turn off desktop manager if it's running"));
+  parser.addOption(desktopOffOption);
+
+  QCommandLineOption desktopPrefOption("desktop-pref", tr("Open desktop preference dialog on the page with the specified name"), tr("NAME"));
+  parser.addOption(desktopPrefOption);
+
+  QCommandLineOption newWindowOption(QStringList() << "n" << "new-window", tr("Open new window"));
+  parser.addOption(newWindowOption);
+
+  QCommandLineOption findFilesOption(QStringList() << "f" << "find-files", tr("Open Find Files utility"));
+  parser.addOption(findFilesOption);
+
+  QCommandLineOption setWallpaperOption(QStringList() << "w" << "set-wallpaper", tr("Set desktop wallpaper from image FILE"), tr("FILE"));
+  parser.addOption(setWallpaperOption);
+
+  // don't translate list of modes in description, please
+  QCommandLineOption wallpaperModeOption("wallpaper-mode", tr("Set mode of desktop wallpaper. MODE=(color|stretch|fit|center|tile)"), tr("MODE"));
+  parser.addOption(wallpaperModeOption);
+
+  QCommandLineOption showPrefOption("show-pref", tr("Open Preferences dialog on the page with the specified name"), tr("NAME"));
+  parser.addOption(showPrefOption);
+
+  parser.addPositionalArgument("files", tr("Files or directories to open"), tr("[FILE1, FILE2,...]"));
+
+  parser.process(arguments());
 
   if(isPrimaryInstance) {
     qDebug("isPrimaryInstance");
 
-    if(daemon_mode)
+    if(parser.isSet(daemonOption))
       daemonMode_ = true;
-    if(profile)
-      profileName_ = profile;
+    if(parser.isSet(profileOption))
+      profileName_ = parser.value(profileOption);
 
     // load settings
     settings_.load(profileName_);
 
     // desktop icon management
-    if(desktop) {
+    if(parser.isSet(desktopOption)) {
       desktopManager(true);
       keepRunning = true;
     }
-    else if(desktop_off)
+    else if(parser.isSet(desktopOffOption))
       desktopManager(false);
 
-    if(desktop_pref) { // desktop preference dialog
-      desktopPrefrences(desktop_pref);
+    if(parser.isSet(desktopPrefOption)) { // desktop preference dialog
+      desktopPrefrences(parser.value(desktopPrefOption));
       keepRunning = true;
     }
-    else if(find_files) { // file searching utility
-      QStringList paths;
-      if(file_names) {
-        for(char** filename = file_names; *filename; ++filename) {
-          QString path(*filename);
-          paths.push_back(path);
-        }
-      }
-      findFiles(paths);
+    else if(parser.isSet(findFilesOption)) { // file searching utility
+      findFiles(parser.positionalArguments());
       keepRunning = true;
     }
-    else if(show_pref) { // preferences dialog
-      preferences(show_pref);
+    else if(parser.isSet(showPrefOption)) { // preferences dialog
+      preferences(parser.value(showPrefOption));
       keepRunning = true;
     }
-    else if(wallpaper || wallpaper_mode) // set wall paper
-      setWallpaper(wallpaper, wallpaper_mode);
+    else if(parser.isSet(setWallpaperOption) || parser.isSet(wallpaperModeOption)) // set wall paper
+      setWallpaper(parser.value(setWallpaperOption), parser.value(wallpaperModeOption));
     else {
-      if(!desktop && !desktop_off) {
-        QStringList paths;
-        if(file_names) {
-          for(char** filename = file_names; *filename; ++filename) {
-            QString path = QString::fromLocal8Bit(*filename);
-            paths.push_back(path);
-          }
-        }
-        else {
+      if(!parser.isSet(desktopOption) && !parser.isSet(desktopOffOption)) {
+        QStringList paths = parser.positionalArguments();
+        if(paths.isEmpty()) {
           // if no path is specified and we're using daemon mode,
           // don't open current working directory
           if(!daemonMode_)
             paths.push_back(QDir::currentPath());
         }
         if(!paths.isEmpty())
-          launchFiles(paths, (bool)new_window);
+          launchFiles(QDir::currentPath(), paths, parser.isSet(newWindowOption));
         keepRunning = true;
       }
     }
@@ -251,60 +251,38 @@ bool Application::parseCommandLineArgs() {
   else {
     QDBusConnection dbus = QDBusConnection::sessionBus();
     QDBusInterface iface(serviceName, "/Application", ifaceName, dbus, this);
-    if(ask_quit) {
+    if(parser.isSet(quitOption)) {
       iface.call("quit");
       return false;
     }
 
-    if(desktop)
+    if(parser.isSet(desktopOption))
       iface.call("desktopManager", true);
-    else if(desktop_off)
+    else if(parser.isSet(desktopOffOption))
       iface.call("desktopManager", false);
 
-    if(desktop_pref) { // desktop preference dialog
-      iface.call("desktopPrefrences", QString(desktop_pref));
+    if(parser.isSet(desktopPrefOption)) { // desktop preference dialog
+      iface.call("desktopPrefrences", parser.value(desktopPrefOption));
     }
-    else if(find_files) { // file searching utility
-      QStringList paths;
-      if(file_names) {
-        for(char** filename = file_names; *filename; ++filename) {
-          QString path(*filename);
-          paths.push_back(path);
-        }
-      }
-      iface.call("findFiles", paths);
+    else if(parser.isSet(findFilesOption)) { // file searching utility
+      iface.call("findFiles", parser.positionalArguments());
     }
-    else if(show_pref) { // preferences dialog
-      iface.call("preferences", QString(show_pref));
+    else if(parser.isSet(showPrefOption)) { // preferences dialog
+      iface.call("preferences", parser.value(showPrefOption));
     }
-    else if(wallpaper || wallpaper_mode) { // set wall paper
-      iface.call("setWallpaper", QString(wallpaper), QString(wallpaper_mode));
+    else if(parser.isSet(setWallpaperOption) || parser.isSet(wallpaperModeOption)) { // set wall paper
+      iface.call("setWallpaper", parser.value(setWallpaperOption), parser.value(wallpaperModeOption));
     }
     else {
-      if(!desktop && !desktop_off) {
-        QStringList paths;
-        if(file_names) {
-          for(char** filename = file_names; *filename; ++filename) {
-            QString path = QString::fromLocal8Bit(*filename);
-            paths.push_back(path);
-          }
-        }
-        else
+      if(!parser.isSet(desktopOption) && !parser.isSet(desktopOffOption)) {
+        QStringList paths = parser.positionalArguments();
+        if(paths.isEmpty()) {
           paths.push_back(QDir::currentPath());
-        // the function requires bool, but new_window is gboolean, the casting is needed.
-        iface.call("launchFiles", paths, (bool)new_window);
+        }
+        iface.call("launchFiles", QDir::currentPath(), paths, parser.isSet(newWindowOption));
       }
     }
   }
-
-  // cleanup
-  g_free(desktop_pref);
-  g_free(show_pref);
-  g_free(wallpaper);
-  g_free(wallpaper_mode);
-  g_free(profile);
-  g_strfreev(file_names);
-
   return keepRunning;
 }
 
@@ -342,6 +320,32 @@ int Application::exec() {
   return QCoreApplication::exec();
 }
 
+
+void Application::onUserDirsChanged()
+{
+  qDebug() << Q_FUNC_INFO;
+  bool file_deleted = !userDirsWatcher_->files().contains(userDirsFile_);
+  if(file_deleted) {
+    // if our config file is already deleted, reinstall a new watcher
+    userDirsWatcher_->addPath(userDirsFile_);
+  }
+
+  const QString d = XdgDir::readDesktopDir();
+  if (d != userDesktopFolder_) {
+    userDesktopFolder_ = d;
+    const QDir dir(d);
+    if (dir.exists()) {
+      const int N = desktopWindows_.size();
+      for(int i = 0; i < N; ++i) {
+        desktopWindows_.at(i)->setDesktopFolder();
+      }
+    } else {
+        qWarning("Application::onUserDirsChanged: %s doesn't exist",
+                    userDesktopFolder_.toUtf8().constData());
+    }
+  }
+}
+
 void Application::onAboutToQuit() {
   qDebug("aboutToQuit");
   settings_.save();
@@ -349,27 +353,15 @@ void Application::onAboutToQuit() {
 
 bool Application::eventFilter(QObject* watched, QEvent* event) {
   if(watched == desktop()) {
-    // qDebug() << "filter" << event;
     switch(event->type()) {
       case QEvent::StyleChange:
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
       case QEvent::ThemeChange:
-#endif
         setStyle(new ProxyStyle());
     };
   }
   return QObject::eventFilter(watched, event);
 }
 
-void Application::commitData(QSessionManager& manager) {
-  qDebug("commitData");
-  // FIXME: where should we write the config file?
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
-#else
-  QApplication::commitData(manager);
-#endif
-}
-
 void Application::onLastWindowClosed() {
 
 }
@@ -384,18 +376,14 @@ void Application::desktopManager(bool enabled) {
   QDesktopWidget* desktopWidget = desktop();
   if(enabled) {
     if(!enableDesktopManager_) {
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
-      installNativeEventFilter(this);
+      // installNativeEventFilter(this);
       Q_FOREACH(QScreen* screen, screens()) {
         connect(screen, &QScreen::virtualGeometryChanged, this, &Application::onVirtualGeometryChanged);
         connect(screen, &QObject::destroyed, this, &Application::onScreenDestroyed);
       }
       connect(this, &QApplication::screenAdded, this, &Application::onScreenAdded);
-#else
-      connect(desktopWidget, SIGNAL(workAreaResized(int)), SLOT(onWorkAreaResized(int)));
-#endif
-      connect(desktopWidget, SIGNAL(resized(int)), SLOT(onScreenResized(int)));
-      connect(desktopWidget, SIGNAL(screenCountChanged(int)), SLOT(onScreenCountChanged(int)));
+      connect(desktopWidget, &QDesktopWidget::resized, this, &Application::onScreenResized);
+      connect(desktopWidget, &QDesktopWidget::screenCountChanged, this, &Application::onScreenCountChanged);
 
       // NOTE: there are two modes
       // When virtual desktop is used (all screens are combined to form a large virtual desktop),
@@ -416,24 +404,20 @@ void Application::desktopManager(bool enabled) {
   }
   else {
     if(enableDesktopManager_) {
-      disconnect(desktopWidget, SIGNAL(resized(int)), this, SLOT(onScreenResized(int)));
-      disconnect(desktopWidget, SIGNAL(screenCountChanged(int)), this, SLOT(onScreenCountChanged(int)));
+      disconnect(desktopWidget, &QDesktopWidget::resized, this, &Application::onScreenResized);
+      disconnect(desktopWidget, &QDesktopWidget::screenCountChanged, this, &Application::onScreenCountChanged);
       int n = desktopWindows_.size();
       for(int i = 0; i < n; ++i) {
         DesktopWindow* window = desktopWindows_.at(i);
         delete window;
       }
       desktopWindows_.clear();
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
       Q_FOREACH(QScreen* screen, screens()) {
         disconnect(screen, &QScreen::virtualGeometryChanged, this, &Application::onVirtualGeometryChanged);
         disconnect(screen, &QObject::destroyed, this, &Application::onScreenDestroyed);
       }
       disconnect(this, &QApplication::screenAdded, this, &Application::onScreenAdded);
-      removeNativeEventFilter(this);
-#else
-      disconnect(desktopWidget, SIGNAL(workAreaResized(int)), this, SLOT(onWorkAreaResized(int)));
-#endif
+      // removeNativeEventFilter(this);
     }
   }
   enableDesktopManager_ = enabled;
@@ -443,6 +427,9 @@ void Application::desktopPrefrences(QString page) {
   // show desktop preference window
   if(!desktopPreferencesDialog_) {
     desktopPreferencesDialog_ = new DesktopPreferencesDialog();
+
+    // Should be used only one time
+    desktopPreferencesDialog_->setEditDesktopFolder(!lxqtRunning_);
   }
   desktopPreferencesDialog_.data()->selectPage(page);
   desktopPreferencesDialog_.data()->show();
@@ -450,20 +437,49 @@ void Application::desktopPrefrences(QString page) {
   desktopPreferencesDialog_.data()->activateWindow();
 }
 
-void Application::findFiles(QStringList paths) {
-  // TODO: add a file searching utility here.
-  qDebug("findFiles");
+void Application::onFindFileAccepted() {
+  Fm::FileSearchDialog* dlg = static_cast<Fm::FileSearchDialog*>(sender());
+  Fm::Path uri = dlg->searchUri();
+  // FIXME: we should be able to open it in an existing window
+  FmPathList* paths = fm_path_list_new();
+  fm_path_list_push_tail(paths, uri.data());
+  Launcher(NULL).launchPaths(NULL, paths);
+  fm_path_list_unref(paths);
 }
 
-void Application::launchFiles(QStringList paths, bool inNewWindow) {
+void Application::findFiles(QStringList paths) {
+  // launch file searching utility.
+  Fm::FileSearchDialog* dlg = new Fm::FileSearchDialog(paths);
+  connect(dlg, &QDialog::accepted, this, &Application::onFindFileAccepted);
+  dlg->setAttribute(Qt::WA_DeleteOnClose);
+  dlg->show();
+}
+
+void Application::launchFiles(QString cwd, QStringList paths, bool inNewWindow) {
   FmPathList* pathList = fm_path_list_new();
+  FmPath* cwd_path = NULL;
   QStringList::iterator it;
-  for(it = paths.begin(); it != paths.end(); ++it) {
-    QString& pathName = *it;
-    FmPath* path = fm_path_new_for_path(pathName.toLocal8Bit().constData());
+  Q_FOREACH(const QString& it, paths) {
+    QByteArray pathName = it.toLocal8Bit();
+    FmPath* path = NULL;
+    if(pathName[0] == '/') // absolute path
+        path = fm_path_new_for_path(pathName.constData());
+    else if(pathName.contains(":/")) // URI
+        path = fm_path_new_for_uri(pathName.constData());
+    else if(pathName == "~") // special case for home dir
+        path = fm_path_ref(fm_path_get_home());
+    else // basename
+    {
+        if(Q_UNLIKELY(!cwd_path))
+            cwd_path = fm_path_new_for_str(cwd.toLocal8Bit().constData());
+        path = fm_path_new_relative(cwd_path, pathName.constData());
+    }
     fm_path_list_push_tail(pathList, path);
     fm_path_unref(path);
   }
+  if(cwd_path)
+      fm_path_unref(cwd_path);
+
   Launcher(NULL).launchPaths(NULL, pathList);
   fm_path_list_unref(pathList);
 }
@@ -558,18 +574,6 @@ void Application::onScreenResized(int num) {
   }
 }
 
-// This slot is for Qt4 only
-void Application::onWorkAreaResized(int num) {
-#if QT_VERSION < QT_VERSION_CHECK(5, 0, 0)
-  if(desktop()->isVirtualDesktop())
-    num = 0; // in virtual desktop mode, we only have one desktop window. that is the first one.
-  DesktopWindow* window = desktopWindows_.at(num);
-  QRect rect = desktop()->availableGeometry(num);
-  qDebug() << "workAreaResized" << num << rect;
-  window->queueRelayout();
-#endif
-}
-
 DesktopWindow* Application::createDesktopWindow(int screenNum) {
   DesktopWindow* window = new DesktopWindow(screenNum);
   if(screenNum == -1) { // one large virtual desktop only
@@ -712,41 +716,31 @@ void Application::onVolumeAdded(GVolumeMonitor* monitor, GVolume* volume, Applic
     pThis->autoMountVolume(volume, true);
 }
 
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
-
+#if 0
 bool Application::nativeEventFilter(const QByteArray & eventType, void * message, long * result) {
   if(eventType == "xcb_generic_event_t") { // XCB event
     // filter all native X11 events (xcb)
     xcb_generic_event_t* generic_event = reinterpret_cast<xcb_generic_event_t*>(message);
     // qDebug("XCB event: %d", generic_event->response_type & ~0x80);
     Q_FOREACH(DesktopWindow * window, desktopWindows_) {
-      window->xcbEvent(generic_event);
     }
   }
   return false;
 }
-
 #endif
 
-// This slot is for Qt 5 onlt, but the stupid Qt moc cannot do conditional compilation
-// so we have to define it for Qt 4 as well.
 void Application::onScreenAdded(QScreen* newScreen) {
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
   if(enableDesktopManager_) {
     connect(newScreen, &QScreen::virtualGeometryChanged, this, &Application::onVirtualGeometryChanged);
     connect(newScreen, &QObject::destroyed, this, &Application::onScreenDestroyed);
   }
-#endif
 }
 
-// This slot is for Qt 5 onlt, but the stupid Qt moc cannot do conditional compilation
-// so we have to define it for Qt 4 as well.
 void Application::onScreenDestroyed(QObject* screenObj) {
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
   // NOTE by PCMan: This is a workaround for Qt 5 bug #40681.
   // With this very dirty workaround, we can fix lxde/lxde-qt bug #204, #205, and #206.
   // Qt 5 has two new regression bugs which breaks lxqt-panel in a multihead environment.
-  // #40681: Regression bug: QWidget::winId() returns old value and QEvent::WinIdChange event is not emitted sometimes. (multihead setup) 
+  // #40681: Regression bug: QWidget::winId() returns old value and QEvent::WinIdChange event is not emitted sometimes. (multihead setup)
   // #40791: Regression: QPlatformWindow, QWindow, and QWidget::winId() are out of sync.
   // Explanations for the workaround:
   // Internally, Qt mantains a list of QScreens and update it when XRandR configuration changes.
@@ -755,7 +749,7 @@ void Application::onScreenDestroyed(QObject* screenObj) {
   // Qt will call QWindow::setScreen(0) on the internal windowHandle() of our panel widget to move it
   // to the primary screen. However, moving a window to a different screen is more than just changing
   // its position. With XRandR, all screens are actually part of the same virtual desktop. However,
-  // this is not the case in other setups, such as Xinerama and moving a window to another screen is 
+  // this is not the case in other setups, such as Xinerama and moving a window to another screen is
   // not possible unless you destroy the widget and create it again for a new screen.
   // Therefore, Qt destroy the widget and re-create it when moving our panel to a new screen.
   // Unfortunately, destroying the window also destroy the child windows embedded into it,
@@ -767,7 +761,6 @@ void Application::onScreenDestroyed(QObject* screenObj) {
   //
   // The workaround is very simple. Just completely destroy the window before Qt has a chance to do
   // QWindow::setScreen() for it. Later, we recreate the window ourselves. So this can bypassing the Qt bugs.
-  QScreen* screen = static_cast<QScreen*>(screenObj);
   if(enableDesktopManager_) {
     bool reloadNeeded = false;
     // FIXME: add workarounds for Qt5 bug #40681 and #40791 here.
@@ -780,13 +773,9 @@ void Application::onScreenDestroyed(QObject* screenObj) {
     if(reloadNeeded)
         QTimer::singleShot(0, this, SLOT(reloadDesktopsAsNeeded()));
   }
-#endif
 }
 
-// This slot is for Qt 5 onlt, but the stupid Qt moc cannot do conditional compilation
-// so we have to define it for Qt 4 as well.
 void Application::reloadDesktopsAsNeeded() {
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
   if(enableDesktopManager_) {
     // workarounds for Qt5 bug #40681 and #40791 here.
     Q_FOREACH(DesktopWindow* desktop, desktopWindows_) {
@@ -797,7 +786,6 @@ void Application::reloadDesktopsAsNeeded() {
       }
     }
   }
-#endif
 }
 
 // This slot is for Qt 5 onlt, but the stupid Qt moc cannot do conditional compilation
@@ -811,13 +799,47 @@ void Application::onVirtualGeometryChanged(const QRect& rect) {
   // However, QScreen provided since Qt5 does not have the bug and
   // virtualGeometryChanged() is emitted correctly when the workAreas changed.
   // So we use it in Qt5.
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
   if(enableDesktopManager_) {
-    QScreen* screeb = static_cast<QScreen*>(sender());
     // qDebug() << "onVirtualGeometryChanged";
     Q_FOREACH(DesktopWindow* desktop, desktopWindows_) {
       desktop->queueRelayout();
     }
   }
-#endif
 }
+
+
+static int sigterm_fd[2];
+
+static void sigtermHandler(int) {
+  char c = 1;
+  ::write(sigterm_fd[0], &c, sizeof(c));
+}
+
+void Application::installSigtermHandler() {
+  if(::socketpair(AF_UNIX, SOCK_STREAM, 0, sigterm_fd) == 0) {
+    QSocketNotifier* notifier = new QSocketNotifier(sigterm_fd[1], QSocketNotifier::Read, this);
+    connect(notifier, &QSocketNotifier::activated, this, &Application::onSigtermNotified);
+
+    struct sigaction action;
+    action.sa_handler = sigtermHandler;
+    ::sigemptyset(&action.sa_mask);
+    action.sa_flags = SA_RESTART;
+    if(::sigaction(SIGTERM, &action, 0) != 0) {
+      qWarning("Couldn't install SIGTERM handler");
+    }
+  } else {
+    qWarning("Couldn't create SIGTERM socketpair");
+  }
+}
+
+void Application::onSigtermNotified() {
+  if (QSocketNotifier* notifier = qobject_cast<QSocketNotifier*>(sender())) {
+    notifier->setEnabled(false);
+    char c;
+    ::read(sigterm_fd[1], &c, sizeof(c));
+    quit();
+    notifier->setEnabled(true);
+  }
+}
+
+} // namespace PCManFM
