@@ -40,25 +40,22 @@
 #include "./application.h"
 #include "mainwindow.h"
 #include "desktopitemdelegate.h"
-#include "foldermenu.h"
-#include "filemenu.h"
-#include "cachedfoldermodel.h"
-#include "folderview_p.h"
-#include "fileoperation.h"
-#include "filepropsdialog.h"
-#include "utilities.h"
+#include <libfm-qt/foldermenu.h>
+#include <libfm-qt/filemenu.h>
+#include <libfm-qt/cachedfoldermodel.h>
+#include <libfm-qt/folderview_p.h>
+#include <libfm-qt/fileoperation.h>
+#include <libfm-qt/filepropsdialog.h>
+#include <libfm-qt/utilities.h>
+#include <libfm-qt/path.h>
+#include "xdgdir.h"
 
-#include <QX11Info> // requires Qt 4 or Qt 5.1
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
+#include <QX11Info>
 #include <QScreen>
 #include <xcb/xcb.h>
 #include <X11/Xlib.h>
-#else
-#include <X11/Xlib.h>
-#include <X11/Xatom.h>
-#endif
 
-using namespace PCManFM;
+namespace PCManFM {
 
 DesktopWindow::DesktopWindow(int screenNum):
   View(Fm::FolderView::IconMode),
@@ -80,7 +77,7 @@ DesktopWindow::DesktopWindow(int screenNum):
   View::setFileLauncher(&fileLauncher_);
 
   listView_ = static_cast<Fm::FolderViewListView*>(childView());
-  listView_->setMovement(QListView::Free);
+  listView_->setMovement(QListView::Snap);
   listView_->setResizeMode(QListView::Adjust);
   listView_->setFlow(QListView::TopToBottom);
 
@@ -92,7 +89,7 @@ DesktopWindow::DesktopWindow(int screenNum):
   if(desktopWidget->isVirtualDesktop() || screenNum_ == desktopWidget->primaryScreen()) {
     loadItemPositions();
     Settings& settings = static_cast<Application* >(qApp)->settings();
-    
+
     model_ = Fm::CachedFolderModel::modelFromPath(fm_path_get_desktop());
     folder_ = reinterpret_cast<FmFolder*>(g_object_ref(model_->folder()));
 
@@ -102,10 +99,10 @@ DesktopWindow::DesktopWindow(int screenNum):
     proxyModel_->sort(Fm::FolderModel::ColumnFileMTime);
     setModel(proxyModel_);
 
-    connect(proxyModel_, SIGNAL(rowsInserted(QModelIndex,int,int)), SLOT(onRowsInserted(QModelIndex,int,int)));
-    connect(proxyModel_, SIGNAL(rowsAboutToBeRemoved(QModelIndex,int,int)), SLOT(onRowsAboutToBeRemoved(QModelIndex,int,int)));
-    connect(proxyModel_, SIGNAL(layoutChanged()), SLOT(onLayoutChanged()));
-    connect(listView_, SIGNAL(indexesMoved(QModelIndexList)), SLOT(onIndexesMoved(QModelIndexList)));
+    connect(proxyModel_, &Fm::ProxyFolderModel::rowsInserted, this, &DesktopWindow::onRowsInserted);
+    connect(proxyModel_, &Fm::ProxyFolderModel::rowsAboutToBeRemoved, this, &DesktopWindow::onRowsAboutToBeRemoved);
+    connect(proxyModel_, &Fm::ProxyFolderModel::layoutChanged, this, &DesktopWindow::onLayoutChanged);
+    connect(listView_, &QListView::indexesMoved, this, &DesktopWindow::onIndexesMoved);
   }
 
   // set our own delegate
@@ -118,19 +115,35 @@ DesktopWindow::DesktopWindow(int screenNum):
   listView_->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
   listView_->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
 
-  connect(this, SIGNAL(openDirRequested(FmPath*, int)), SLOT(onOpenDirRequested(FmPath*, int)));
-  
+  connect(this, &DesktopWindow::openDirRequested, this, &DesktopWindow::onOpenDirRequested);
+
   listView_->installEventFilter(this);
-  
+
   // setup shortcuts
-  new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_X), this, SLOT(onCutActivated())); // cut
-  new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_C), this, SLOT(onCopyActivated())); // copy
-  new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_V), this, SLOT(onPasteActivated())); // paste
-  new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_A), listView_, SLOT(selectAll())); // select all
-  new QShortcut(QKeySequence(Qt::Key_Delete), this, SLOT(onDeleteActivated())); // delete
-  new QShortcut(QKeySequence(Qt::Key_F2), this, SLOT(onRenameActivated())); // rename
-  new QShortcut(QKeySequence(Qt::ALT + Qt::Key_Return), this, SLOT(onFilePropertiesActivated())); // rename
-  new QShortcut(QKeySequence(Qt::SHIFT + Qt::Key_Delete), this, SLOT(onDeleteActivated())); // force delete
+  QShortcut* shortcut;
+  shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_X), this); // cut
+  connect(shortcut, &QShortcut::activated, this, &DesktopWindow::onCutActivated);
+
+  shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_C), this); // copy
+  connect(shortcut, &QShortcut::activated, this, &DesktopWindow::onCopyActivated);
+
+  shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_V), this); // paste
+  connect(shortcut, &QShortcut::activated, this, &DesktopWindow::onPasteActivated);
+
+  shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_A), this); // select all
+  connect(shortcut, &QShortcut::activated, listView_, &QListView::selectAll);
+
+  shortcut = new QShortcut(QKeySequence(Qt::Key_Delete), this); // delete
+  connect(shortcut, &QShortcut::activated, this, &DesktopWindow::onDeleteActivated);
+
+  shortcut = new QShortcut(QKeySequence(Qt::Key_F2), this); // rename
+  connect(shortcut, &QShortcut::activated, this, &DesktopWindow::onRenameActivated);
+
+  shortcut = new QShortcut(QKeySequence(Qt::ALT + Qt::Key_Return), this); // rename
+  connect(shortcut, &QShortcut::activated, this, &DesktopWindow::onFilePropertiesActivated);
+
+  shortcut = new QShortcut(QKeySequence(Qt::SHIFT + Qt::Key_Delete), this); // force delete
+  connect(shortcut, &QShortcut::activated, this, &DesktopWindow::onDeleteActivated);
 }
 
 DesktopWindow::~DesktopWindow() {
@@ -185,6 +198,12 @@ void DesktopWindow::resizeEvent(QResizeEvent* event) {
   queueRelayout(100); // Qt use a 100 msec delay for relayout internally so we use it, too.
 }
 
+void DesktopWindow::setDesktopFolder() {
+  FmPath *path = fm_path_new_for_path(XdgDir::readDesktopDir().toStdString().c_str());
+  model_ = Fm::CachedFolderModel::modelFromPath(path);
+  proxyModel_->setSourceModel(model_);
+}
+
 void DesktopWindow::setWallpaperFile(QString filename) {
   wallpaperFile_ = filename;
 }
@@ -320,6 +339,7 @@ void DesktopWindow::updateWallpaper() {
 }
 
 void DesktopWindow::updateFromSettings(Settings& settings) {
+  setDesktopFolder();
   setWallpaperFile(settings.wallpaper());
   setWallpaperMode(settings.wallpaperMode());
   setFont(settings.desktopFont());
@@ -357,7 +377,7 @@ void DesktopWindow::prepareFileMenu(Fm::FileMenu* menu) {
       action->setChecked(true);
     }
   }
-  connect(action, SIGNAL(toggled(bool)), SLOT(onStickToCurrentPos(bool)));
+  connect(action, &QAction::toggled, this, &DesktopWindow::onStickToCurrentPos);
 }
 
 void DesktopWindow::prepareFolderMenu(Fm::FolderMenu* menu) {
@@ -366,100 +386,9 @@ void DesktopWindow::prepareFolderMenu(Fm::FolderMenu* menu) {
   menu->removeAction(menu->propertiesAction());
   // add an action for desktop preferences instead
   QAction* action = menu->addAction(tr("Desktop Preferences"));
-  connect(action, SIGNAL(triggered(bool)), SLOT(onDesktopPreferences()));
+  connect(action, &QAction::triggered, this, &DesktopWindow::onDesktopPreferences);
 }
 
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0) // Qt 5
-
-void DesktopWindow::xcbEvent(xcb_generic_event_t* generic_event) {
-  int event_type = generic_event->response_type & ~0x80;
-  if(showWmMenu_) {
-    // If we want to show the desktop menus provided by the window manager instead of ours,
-    // we have to forward the mouse events we received to the root window.
-    switch(event_type) {
-      case XCB_BUTTON_PRESS: {
-        xcb_button_press_event_t* event = reinterpret_cast<xcb_button_press_event_t*>(generic_event);
-        if(event->event == effectiveWinId()) {
-          // check if the user click on blank area
-          QModelIndex index = listView_->indexAt(QPoint(event->event_x, event->event_y));
-          if(!index.isValid()) {
-            xcb_ungrab_pointer(QX11Info::connection(), event->time);
-            // forward the event to the root window
-            xcb_button_press_event_t event2 = *event;
-            WId root = QX11Info::appRootWindow(QX11Info::appScreen());
-            event2.event = root;
-            xcb_send_event(QX11Info::connection(), 0, root, XCB_EVENT_MASK_BUTTON_PRESS, (char*)&event2);
-          }
-        }
-        break;
-      }
-      case XCB_BUTTON_RELEASE: {
-        xcb_button_release_event_t* event = reinterpret_cast<xcb_button_release_event_t*>(generic_event);
-        if(event->event == effectiveWinId()) {
-          // check if the user click on blank area
-          QModelIndex index = listView_->indexAt(QPoint(event->event_x, event->event_y));
-          if(!index.isValid()) {
-            // forward the event to the root window
-            xcb_button_release_event_t event2 = *event;
-            WId root = QX11Info::appRootWindow(QX11Info::appScreen());
-            event2.event = root;
-            xcb_send_event(QX11Info::connection(), 0, root, XCB_EVENT_MASK_BUTTON_RELEASE, (char*)&event2);
-          }
-        }
-        break;
-      }
-      default:
-        break;
-    }
-  }
-}
-
-#else // Qt 4
-
-bool DesktopWindow::x11Event(XEvent * event) {
-  // FIXME: support Qt4
-  if(showWmMenu_) {
-    // filter all native X11 events (using Xlib)
-    switch(event->type) {
-      case ButtonPress: {
-        if(event->xany.window == effectiveWinId()) {
-          // check if the user click on blank area
-          QModelIndex index = listView_->indexAt(QPoint(event->xbutton.x, event->xbutton.y));
-          if(!index.isValid()) {
-            XUngrabPointer(QX11Info::display(), event->xbutton.time);
-            // forward the event to the root window
-            XButtonEvent event2 = event->xbutton;
-            WId root = QX11Info::appRootWindow(QX11Info::appScreen());
-            event2.window = root;
-            XSendEvent(QX11Info::display(), root, False, ButtonPressMask|ButtonReleaseMask, (XEvent*)&event2);
-          }
-        }
-        break;
-      }
-      case ButtonRelease: {
-        if(event->xany.window == effectiveWinId()) {
-          // check if the user click on blank area
-          QModelIndex index = listView_->indexAt(QPoint(event->xbutton.x, event->xbutton.y));
-          if(!index.isValid()) {
-            // forward the event to the root window
-            XButtonEvent event2 = event->xbutton;
-            WId root = QX11Info::appRootWindow(QX11Info::appScreen());
-            event2.window = root;
-            XSendEvent(QX11Info::display(), root, False, ButtonPressMask|ButtonReleaseMask, (XEvent*)&event2);
-          }
-        }
-        break;
-      }
-      default: {
-        break;
-      }
-    }
-  }
-  return QWidget::x11Event(event);
-}
-
-#endif
-
 void DesktopWindow::onDesktopPreferences() {
   static_cast<Application* >(qApp)->desktopPrefrences(QString());
 }
@@ -469,18 +398,24 @@ void DesktopWindow::onRowsInserted(const QModelIndex& parent, int start, int end
 }
 
 void DesktopWindow::onRowsAboutToBeRemoved(const QModelIndex& parent, int start, int end) {
+  Q_UNUSED(parent);
+  Q_UNUSED(start);
+  Q_UNUSED(end);
   if(!customItemPos_.isEmpty()) {
     // also delete stored custom item positions for the items currently being removed.
-    bool changed = false;
-    for(int row = start; row <= end ;++row) {
-      QModelIndex index = parent.child(row, 0);
-      FmFileInfo* file = proxyModel_->fileInfoFromIndex(index);
-      if(file) { // remove custom position for the item
-        if(customItemPos_.remove(fm_file_info_get_name(file)))
-          changed = true;
-      }
+    // Here we can't rely on ProxyFolderModel::fileInfoFromIndex() because, although rows
+    // aren't removed yet, files are already removed.
+    QHash<QByteArray, QPoint> _customItemPos = customItemPos_;
+    char* dektopPath = fm_path_to_str(fm_path_get_desktop());
+    QString desktopDir = QString(dektopPath) + QString("/");
+    g_free(dektopPath);
+    QHash<QByteArray, QPoint>::iterator it;
+    for(it = _customItemPos.begin(); it != _customItemPos.end(); ++it) {
+      const QByteArray& name = it.key();
+      if(!QFile::exists(desktopDir + QString::fromUtf8(name, name.length())))
+        customItemPos_.remove(it.key());
     }
-    if(changed)
+    if(customItemPos_ != _customItemPos)
       saveItemPositions();
   }
   queueRelayout();
@@ -493,7 +428,7 @@ void DesktopWindow::onLayoutChanged() {
 void DesktopWindow::onIndexesMoved(const QModelIndexList& indexes) {
   // remember the custom position for the items
   Q_FOREACH(const QModelIndex& index, indexes) {
-    // Under some circumstances, Qt might emit indexMoved for 
+    // Under some circumstances, Qt might emit indexMoved for
     // every single cells in the same row. (when QAbstractItemView::SelectItems is set)
     // So indexes list may contain several indixes for the same row.
     // Since we only care about rows, not individual cells,
@@ -501,9 +436,17 @@ void DesktopWindow::onIndexesMoved(const QModelIndexList& indexes) {
     if(index.column() == 0) {
       FmFileInfo* file = proxyModel_->fileInfoFromIndex(index);
       QRect itemRect = listView_->rectForIndex(index);
-      QByteArray name = fm_file_info_get_name(file);
-      customItemPos_[name] = itemRect.topLeft();
-      // qDebug() << "indexMoved:" << name << index << itemRect;
+      QPoint tl = itemRect.topLeft();
+      QRect workArea = qApp->desktop()->availableGeometry(screenNum_);
+      workArea.adjust(12, 12, -12, -12);
+      if(customItemPos_.keys(tl).isEmpty() // don't put items on each other
+         && tl.x() >= workArea.x() && tl.y() >= workArea.y()
+         && tl.x() + listView_->gridSize().width() <= workArea.right()
+         && tl.y() + listView_->gridSize().height() <= workArea.bottom()) {
+        QByteArray name = fm_file_info_get_name(file);
+        customItemPos_[name] = tl;
+        // qDebug() << "indexMoved:" << name << index << itemRect;
+      }
     }
   }
   saveItemPositions();
@@ -513,13 +456,14 @@ void DesktopWindow::onIndexesMoved(const QModelIndexList& indexes) {
 // QListView does item layout in a very inflexible way, so let's do our custom layout again.
 // FIXME: this is very inefficient, but due to the design flaw of QListView, this is currently the only workaround.
 void DesktopWindow::relayoutItems() {
+  loadItemPositions(); // something may have changed
   // qDebug("relayoutItems()");
   if(relayoutTimer_) {
     // this slot might be called from the timer, so we cannot delete it directly here.
     relayoutTimer_->deleteLater();
     relayoutTimer_ = NULL;
   }
-  
+
   QDesktopWidget* desktop = qApp->desktop();
   int screen = 0;
   int row = 0;
@@ -539,12 +483,14 @@ void DesktopWindow::relayoutItems() {
     QPoint pos = workArea.topLeft();
     for(; row < rowCount; ++row) {
       QModelIndex index = proxyModel_->index(row, 0);
+      int itemWidth = delegate_->sizeHint(listView_->getViewOptions(), index).width();
       FmFileInfo* file = proxyModel_->fileInfoFromIndex(index);
       QByteArray name = fm_file_info_get_name(file);
       QHash<QByteArray, QPoint>::iterator it = customItemPos_.find(name);
       if(it != customItemPos_.end()) { // the item has a custom position
         QPoint customPos = *it;
-        listView_->setPositionForIndex(customPos, index);
+        // center the contents vertically
+        listView_->setPositionForIndex(customPos + QPoint((grid.width() - itemWidth) / 2, 0), index);
         // qDebug() << "set custom pos:" << name << row << index << customPos;
         continue;
       }
@@ -561,7 +507,8 @@ void DesktopWindow::relayoutItems() {
         --row;
       }
       else {
-        listView_->setPositionForIndex(pos, index);
+        // center the contents vertically
+        listView_->setPositionForIndex(pos + QPoint((grid.width() - itemWidth) / 2, 0), index);
         // qDebug() << "set pos" << name << row << index << pos;
       }
       // move to next cell in the column
@@ -588,14 +535,46 @@ void DesktopWindow::relayoutItems() {
 
 void DesktopWindow::loadItemPositions() {
   // load custom item positions
+  customItemPos_.clear();
   Settings& settings = static_cast<Application*>(qApp)->settings();
   QString configFile = QString("%1/desktop-items-%2.conf").arg(settings.profileDir(settings.profileName())).arg(screenNum_);
   QSettings file(configFile, QSettings::IniFormat);
+  QSize grid = listView_->gridSize();
+  QRect workArea = qApp->desktop()->availableGeometry(screenNum_);
+  workArea.adjust(12, 12, -12, -12);
+  char* dektopPath = fm_path_to_str(fm_path_get_desktop());
+  QString desktopDir = QString(dektopPath) + QString("/");
+  g_free(dektopPath);
   Q_FOREACH(const QString& name, file.childGroups()) {
+    if(!QFile::exists(desktopDir + name.toUtf8())) {
+      // the file may have been removed from outside LXQT
+      continue;
+    }
     file.beginGroup(name);
     QVariant var = file.value("pos");
-    if(var.isValid())
-      customItemPos_[name.toUtf8()] = var.toPoint();
+    if(var.isValid()) {
+      QPoint customPos = var.toPoint();
+      if (customPos.x() >= workArea.x() && customPos.y() >= workArea.y()
+          && customPos.x() + listView_->gridSize().width() <= workArea.right()
+          && customPos.y() + listView_->gridSize().height() <= workArea.bottom())
+      {
+        // correct positions that are't aligned to the grid
+        qreal w = qAbs((qreal)customPos.x() - (qreal)workArea.x())
+                  / (qreal)(grid.width() + listView_->spacing());
+        qreal h = qAbs(customPos.y() - (qreal)workArea.y())
+                  / (qreal)(grid.height() + listView_->spacing());
+        customPos.setX(workArea.x() + qRound(w) * (grid.width() + listView_->spacing()));
+        customPos.setY(workArea.y() + qRound(h) * (grid.height() + listView_->spacing()));
+        while(customItemPos_.values().contains(customPos)) {
+          customPos.setY(customPos.y() + grid.height() + listView_->spacing());
+          if(customPos.y() + grid.height() > workArea.bottom()) {
+            customPos.setX(customPos.x() + grid.width() + listView_->spacing());
+            customPos.setY(workArea.top());
+          }
+        }
+        customItemPos_[name.toUtf8()] = customPos;
+      }
+    }
     file.endGroup();
   }
 }
@@ -649,7 +628,7 @@ void DesktopWindow::queueRelayout(int delay) {
   if(!relayoutTimer_) {
     relayoutTimer_ = new QTimer();
     relayoutTimer_->setSingleShot(true);
-    connect(relayoutTimer_, SIGNAL(timeout()), SLOT(relayoutItems()));
+    connect(relayoutTimer_, &QTimer::timeout, this, &DesktopWindow::relayoutItems);
     relayoutTimer_->start(delay);
   }
 }
@@ -679,7 +658,7 @@ void DesktopWindow::onDeleteActivated() {
     Settings& settings = static_cast<Application*>(qApp)->settings();
     bool shiftPressed = (qApp->keyboardModifiers() & Qt::ShiftModifier ? true : false);
     if(settings.useTrash() && !shiftPressed)
-      Fm::FileOperation::trashFiles(paths, settings.confirmDelete());
+      Fm::FileOperation::trashFiles(paths, settings.confirmTrash());
     else
       Fm::FileOperation::deleteFiles(paths, settings.confirmDelete());
     fm_path_list_unref(paths);
@@ -703,6 +682,69 @@ void DesktopWindow::onFilePropertiesActivated() {
   }
 }
 
+static void forwardMouseEventToRoot(QMouseEvent* event) {
+  xcb_ungrab_pointer(QX11Info::connection(), event->timestamp());
+  // forward the event to the root window
+  xcb_button_press_event_t xcb_event;
+  uint32_t mask = 0;
+  xcb_event.state = 0;
+  switch(event->type()) {
+  case QEvent::MouseButtonPress:
+    xcb_event.response_type = XCB_BUTTON_PRESS;
+    mask = XCB_EVENT_MASK_BUTTON_PRESS;
+    break;
+  case QEvent::MouseButtonRelease:
+      xcb_event.response_type = XCB_BUTTON_RELEASE;
+      mask = XCB_EVENT_MASK_BUTTON_RELEASE;
+    break;
+  default:
+    return;
+  }
+
+  // convert Qt button to XCB button
+  switch(event->button()) {
+  case Qt::LeftButton:
+    xcb_event.detail = 1;
+    xcb_event.state |= XCB_BUTTON_MASK_1;
+    break;
+  case Qt::MiddleButton:
+    xcb_event.detail = 2;
+    xcb_event.state |= XCB_BUTTON_MASK_2;
+    break;
+  case Qt::RightButton:
+    xcb_event.detail = 3;
+    xcb_event.state |= XCB_BUTTON_MASK_3;
+    break;
+  default:
+    xcb_event.detail = 0;
+  }
+
+  // convert Qt modifiers to XCB states
+  if(event->modifiers() & Qt::ShiftModifier)
+    xcb_event.state |= XCB_MOD_MASK_SHIFT;
+  if(event->modifiers() & Qt::ControlModifier)
+    xcb_event.state |= XCB_MOD_MASK_SHIFT;
+  if(event->modifiers() & Qt::AltModifier)
+    xcb_event.state |= XCB_MOD_MASK_1;
+
+  xcb_event.sequence = 0;
+  xcb_event.time = event->timestamp();
+
+  WId root = QX11Info::appRootWindow(QX11Info::appScreen());
+  xcb_event.event = root;
+  xcb_event.root = root;
+  xcb_event.child = 0;
+
+  xcb_event.root_x = event->globalX();
+  xcb_event.root_y = event->globalY();
+  xcb_event.event_x = event->x();
+  xcb_event.event_y = event->y();
+  xcb_event.same_screen = 1;
+
+  xcb_send_event(QX11Info::connection(), 0, root, mask, (char*)&xcb_event);
+  xcb_flush(QX11Info::connection());
+}
+
 bool DesktopWindow::event(QEvent* event)
 {
   switch(event->type()) {
@@ -711,7 +753,6 @@ bool DesktopWindow::event(QEvent* event)
     if(effectiveWinId() == 0)
       break;
     // set freedesktop.org EWMH hints properly
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
     if(QX11Info::isPlatformX11() && QX11Info::connection()) {
       xcb_connection_t* con = QX11Info::connection();
       const char* atom_name = "_NET_WM_WINDOW_TYPE_DESKTOP";
@@ -721,10 +762,6 @@ bool DesktopWindow::event(QEvent* event)
       xcb_atom_t XA_ATOM = 4;
       xcb_change_property(con, XCB_PROP_MODE_REPLACE, effectiveWinId(), prop_atom, XA_ATOM, 32, 1, &atom);
     }
-#else
-    Atom atom = XInternAtom(QX11Info::display(), "_NET_WM_WINDOW_TYPE_DESKTOP", False);
-    XChangeProperty(QX11Info::display(), effectiveWinId(), XInternAtom(QX11Info::display(), "_NET_WM_WINDOW_TYPE", False), XA_ATOM, 32, PropModeReplace, (uchar*)&atom, 1);
-#endif
     break;
   }
 #undef FontChange // FontChange is defined in the headers of XLib and clashes with Qt, let's undefine it.
@@ -732,7 +769,11 @@ bool DesktopWindow::event(QEvent* event)
   case QEvent::FontChange:
     queueRelayout();
     break;
+
+  default:
+      break;
   }
+
   return QWidget::event(event);
 }
 
@@ -750,6 +791,25 @@ bool DesktopWindow::eventFilter(QObject * watched, QEvent * event) {
       break;
     }
   }
+  else if(watched == listView_->viewport()) {
+    switch(event->type()) {
+    case QEvent::MouseButtonPress:
+    case QEvent::MouseButtonRelease:
+        if(showWmMenu_) {
+          QMouseEvent* e = static_cast<QMouseEvent*>(event);
+          // If we want to show the desktop menus provided by the window manager instead of ours,
+          // we have to forward the mouse events we received to the root window.
+          // check if the user click on blank area
+          QModelIndex index = listView_->indexAt(e->pos());
+          if(!index.isValid() && e->button() != Qt::LeftButton) {
+            forwardMouseEventToRoot(e);
+          }
+      }
+      break;
+    default:
+      break;
+    }
+  }
   return false;
 }
 
@@ -778,6 +838,10 @@ void DesktopWindow::childDropEvent(QDropEvent* e) {
     Fm::FolderView::childDropEvent(e);
 }
 
+void DesktopWindow::closeEvent(QCloseEvent *event) {
+  // prevent the desktop window from being closed.
+  event->ignore();
+}
 
 void DesktopWindow::setScreenNum(int num) {
   if(screenNum_ != num) {
@@ -785,3 +849,5 @@ void DesktopWindow::setScreenNum(int num) {
     queueRelayout();
   }
 }
+
+} // namespace PCManFM
