@@ -21,10 +21,10 @@
 #ifndef PCMANFM_LAUNCHER_H
 #define PCMANFM_LAUNCHER_H
 
-#include "filelauncher.h"
+#include <libfm-qt/filelauncher.h>
 
 namespace PCManFM {
-  
+
 class MainWindow;
 
 class Launcher : public Fm::FileLauncher {
