@@ -32,16 +32,16 @@
 #include <QDebug>
 
 #include "tabpage.h"
-#include "filelauncher.h"
-#include "filemenu.h"
-#include "bookmarkaction.h"
-#include "fileoperation.h"
-#include "utilities.h"
-#include "filepropsdialog.h"
-#include "pathedit.h"
+#include "launcher.h"
+#include <libfm-qt/filemenu.h>
+#include <libfm-qt/bookmarkaction.h>
+#include <libfm-qt/fileoperation.h>
+#include <libfm-qt/utilities.h>
+#include <libfm-qt/filepropsdialog.h>
+#include <libfm-qt/pathedit.h>
 #include "ui_about.h"
 #include "application.h"
-#include "../libfm-qt/path.h"
+#include <libfm-qt/path.h>
 
 // #include "qmodeltest/modeltest.h"
 
@@ -51,7 +51,8 @@ namespace PCManFM {
 
 MainWindow::MainWindow(FmPath* path):
   QMainWindow(),
-  fileLauncher_(this) {
+  fileLauncher_(this),
+  rightClickIndex(-1) {
 
   Settings& settings = static_cast<Application*>(qApp)->settings();
   setAttribute(Qt::WA_DeleteOnClose);
@@ -59,38 +60,70 @@ MainWindow::MainWindow(FmPath* path):
   ui.setupUi(this);
 
   // hide menu items that are not usable
-  if(!isUriSchemeSupported("computer"))
-    ui.actionComputer->setVisible(false);
-  if(!isUriSchemeSupported("trash"))
+  //if(!uriExists("computer:///"))
+  //  ui.actionComputer->setVisible(false);
+  if(!settings.supportTrash())
     ui.actionTrash->setVisible(false);
-  if(!isUriSchemeSupported("network"))
-    ui.actionNetwork->setVisible(false);
+
+  // FIXME: add an option to hide network:///
+  // We cannot use uriExists() here since calling this on "network:///"
+  // is very slow and blocking.
+  //if(!uriExists("network:///"))
+  //  ui.actionNetwork->setVisible(false);
 
   // add a context menu for showing browse history to back and forward buttons
   QToolButton* forwardButton = static_cast<QToolButton*>(ui.toolBar->widgetForAction(ui.actionGoForward));
   forwardButton->setContextMenuPolicy(Qt::CustomContextMenu);
-  connect(forwardButton, SIGNAL(customContextMenuRequested(QPoint)), SLOT(onBackForwardContextMenu(QPoint)));
+  connect(forwardButton, &QToolButton::customContextMenuRequested, this, &MainWindow::onBackForwardContextMenu);
   QToolButton* backButton = static_cast<QToolButton*>(ui.toolBar->widgetForAction(ui.actionGoBack));
   backButton->setContextMenuPolicy(Qt::CustomContextMenu);
-  connect(backButton, SIGNAL(customContextMenuRequested(QPoint)), SLOT(onBackForwardContextMenu(QPoint)));
+  connect(backButton, &QToolButton::customContextMenuRequested, this, &MainWindow::onBackForwardContextMenu);
 
   // tabbed browsing interface
   ui.tabBar->setDocumentMode(true);
   ui.tabBar->setTabsClosable(true);
   ui.tabBar->setElideMode(Qt::ElideRight);
   ui.tabBar->setExpanding(false);
-  connect(ui.tabBar, SIGNAL(currentChanged(int)), SLOT(onTabBarCurrentChanged(int)));
-  connect(ui.tabBar, SIGNAL(tabCloseRequested(int)), SLOT(onTabBarCloseRequested(int)));
-  connect(ui.stackedWidget, SIGNAL(widgetRemoved(int)), SLOT(onStackedWidgetWidgetRemoved(int)));
+  ui.tabBar->setMovable(true); // reorder the tabs by dragging
+
+#if QT_VERSION >= QT_VERSION_CHECK(5, 4, 0)
+  // switch to the tab under the cursor during dnd.
+  ui.tabBar->setChangeCurrentOnDrag(true);
+  ui.tabBar->setAcceptDrops(true);
+#endif
+
+  ui.tabBar->setContextMenuPolicy(Qt::CustomContextMenu);
+  connect(ui.actionCloseRight, &QAction::triggered, this, &MainWindow::closeRightTabs);
+  connect(ui.actionCloseLeft, &QAction::triggered, this, &MainWindow::closeLeftTabs);
+  connect(ui.actionCloseOther, &QAction::triggered, this, &MainWindow::closeOtherTabs);
+
+  connect(ui.tabBar, &QTabBar::currentChanged, this, &MainWindow::onTabBarCurrentChanged);
+  connect(ui.tabBar, &QTabBar::tabCloseRequested, this, &MainWindow::onTabBarCloseRequested);
+  connect(ui.tabBar, &QTabBar::tabMoved, this, &MainWindow::onTabBarTabMoved);
+  connect(ui.tabBar, &QTabBar::customContextMenuRequested, this, &MainWindow::tabContextMenu);
+  connect(ui.stackedWidget, &QStackedWidget::widgetRemoved, this, &MainWindow::onStackedWidgetWidgetRemoved);
+
+  // FIXME: should we make the filter bar a per-view configuration?
+  ui.filterBar->setVisible(settings.showFilter());
+  ui.actionFilter->setChecked(settings.showFilter());
+  connect(ui.filterBar, &QLineEdit::textChanged, this, &MainWindow::onFilterStringChanged);
 
   // side pane
   ui.sidePane->setIconSize(QSize(settings.sidePaneIconSize(), settings.sidePaneIconSize()));
-  ui.sidePane->setMode(Fm::SidePane::ModePlaces);
-  connect(ui.sidePane, SIGNAL(chdirRequested(int, FmPath*)), SLOT(onSidePaneChdirRequested(int, FmPath*)));
+  ui.sidePane->setMode(settings.sidePaneMode());
+  connect(ui.sidePane, &Fm::SidePane::chdirRequested, this, &MainWindow::onSidePaneChdirRequested);
+  connect(ui.sidePane, &Fm::SidePane::openFolderInNewWindowRequested, this, &MainWindow::onSidePaneOpenFolderInNewWindowRequested);
+  connect(ui.sidePane, &Fm::SidePane::openFolderInNewTabRequested, this, &MainWindow::onSidePaneOpenFolderInNewTabRequested);
+  connect(ui.sidePane, &Fm::SidePane::openFolderInTerminalRequested, this, &MainWindow::onSidePaneOpenFolderInTerminalRequested);
+  connect(ui.sidePane, &Fm::SidePane::createNewFolderRequested, this, &MainWindow::onSidePaneCreateNewFolderRequested);
+  connect(ui.sidePane, &Fm::SidePane::modeChanged, this, &MainWindow::onSidePaneModeChanged);
+
+  // detect change of splitter position
+  connect(ui.splitter, &QSplitter::splitterMoved, this, &MainWindow::onSplitterMoved);
 
   // path bar
   pathEntry = new Fm::PathEdit(this);
-  connect(pathEntry, SIGNAL(returnPressed()), SLOT(onPathEntryReturnPressed()));
+  connect(pathEntry, &Fm::PathEdit::returnPressed, this, &MainWindow::onPathEntryReturnPressed);
   ui.toolBar->insertWidget(ui.actionGo, pathEntry);
 
   // add filesystem info to status bar
@@ -100,7 +133,7 @@ MainWindow::MainWindow(FmPath* path):
   // setup the splitter
   ui.splitter->setStretchFactor(1, 1); // only the right pane can be stretched
   QList<int> sizes;
-  sizes.append(150);
+  sizes.append(settings.splitterPos());
   sizes.append(300);
   ui.splitter->setSizes(sizes);
 
@@ -132,34 +165,72 @@ MainWindow::MainWindow(FmPath* path):
   group->addAction(ui.actionDescending);
 
   // create shortcuts
-  new QShortcut(Qt::CTRL + Qt::Key_L, pathEntry, SLOT(setFocus()));
-  new QShortcut(Qt::ALT + Qt::Key_D, pathEntry, SLOT(setFocus()));
+  QShortcut* shortcut;
+  shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_L), this);
+  connect(shortcut, &QShortcut::activated, pathEntry, static_cast<void (QWidget::*)()>(&Fm::PathEdit::setFocus));
 
-  new QShortcut(Qt::CTRL + Qt::Key_Tab, this, SLOT(onShortcutNextTab()));
-  new QShortcut(Qt::CTRL + Qt::SHIFT + Qt::Key_Tab, this, SLOT(onShortcutPrevTab()));
+  shortcut = new QShortcut(Qt::ALT + Qt::Key_D, this);
+  connect(shortcut, &QShortcut::activated, pathEntry, static_cast<void (QWidget::*)()>(&QWidget::setFocus));
+
+  shortcut = new QShortcut(Qt::CTRL + Qt::Key_Tab, this);
+  connect(shortcut, &QShortcut::activated, this, &MainWindow::onShortcutNextTab);
+
+  shortcut = new QShortcut(Qt::CTRL + Qt::SHIFT + Qt::Key_Tab, this);
+  connect(shortcut, &QShortcut::activated, this, &MainWindow::onShortcutPrevTab);
+
+  // Add Ctrl+PgUp and Ctrl+PgDown as well, because they are common in Firefox
+  // , Opera, Google Chromium/Google Chrome and most other tab-using
+  // applications.
+  shortcut = new QShortcut(Qt::CTRL + Qt::Key_PageDown, this);
+  connect(shortcut, &QShortcut::activated, this, &MainWindow::onShortcutNextTab);
+
+  shortcut = new QShortcut(Qt::CTRL + Qt::Key_PageUp, this);
+  connect(shortcut, &QShortcut::activated, this, &MainWindow::onShortcutPrevTab);
 
   int i;
   for(i = 0; i < 10; ++i) {
-    new QShortcut(QKeySequence(Qt::ALT + Qt::Key_0 + i), this, SLOT(onShortcutJumpToTab()));
-    new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_0 + i), this, SLOT(onShortcutJumpToTab()));
+    shortcut = new QShortcut(QKeySequence(Qt::ALT + Qt::Key_0 + i), this);
+    connect(shortcut, &QShortcut::activated, this, &MainWindow::onShortcutJumpToTab);
+
+    shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_0 + i), this);
+    connect(shortcut, &QShortcut::activated, this, &MainWindow::onShortcutJumpToTab);
   }
 
-  new QShortcut(QKeySequence(Qt::Key_Backspace), this, SLOT(on_actionGoUp_triggered()));
-  new QShortcut(QKeySequence(Qt::SHIFT + Qt::Key_Delete), this, SLOT(on_actionDelete_triggered()));
+  shortcut = new QShortcut(QKeySequence(Qt::Key_Backspace), this);
+  connect(shortcut, &QShortcut::activated, this, &MainWindow::on_actionGoUp_triggered);
+
+  shortcut = new QShortcut(QKeySequence(Qt::SHIFT + Qt::Key_Delete), this);
+  connect(shortcut, &QShortcut::activated, this, &MainWindow::on_actionDelete_triggered);
+
+  if(QToolButton* clearButton = ui.filterBar->findChild<QToolButton*>()) {
+    clearButton->setToolTip(tr("Clear text (Ctrl+K)"));
+    shortcut = new QShortcut(Qt::CTRL + Qt::Key_K, this);
+    connect(shortcut, &QShortcut::activated, ui.filterBar, &QLineEdit::clear);
+  }
 
   if(path)
     addTab(path);
+
+  // size from settings
+  if(settings.rememberWindowSize()) {
+    resize(settings.windowWidth(), settings.windowHeight());
+    if(settings.windowMaximized())
+      setWindowState(windowState() | Qt::WindowMaximized);
+  }
 }
 
 MainWindow::~MainWindow() {
-  if(bookmarks)
+  if(bookmarks) {
+    g_signal_handlers_disconnect_by_func(bookmarks, (gpointer)G_CALLBACK(onBookmarksChanged), this);
     g_object_unref(bookmarks);
+  }
 }
 
 void MainWindow::chdir(FmPath* path) {
   TabPage* page = currentPage();
 
   if(page) {
+    ui.filterBar->clear();
     page->chdir(path, true);
     updateUIForCurrentPage();
   }
@@ -172,10 +243,12 @@ void MainWindow::addTab(FmPath* path) {
   TabPage* newPage = new TabPage(path, this);
   newPage->setFileLauncher(&fileLauncher_);
   int index = ui.stackedWidget->addWidget(newPage);
-  connect(newPage, SIGNAL(titleChanged(QString)), SLOT(onTabPageTitleChanged(QString)));
-  connect(newPage, SIGNAL(statusChanged(int, QString)), SLOT(onTabPageStatusChanged(int, QString)));
-  connect(newPage, SIGNAL(openDirRequested(FmPath*, int)), SLOT(onTabPageOpenDirRequested(FmPath*, int)));
-  connect(newPage, SIGNAL(sortFilterChanged()), SLOT(onTabPageSortFilterChanged()));
+  connect(newPage, &TabPage::titleChanged, this, &MainWindow::onTabPageTitleChanged);
+  connect(newPage, &TabPage::statusChanged, this, &MainWindow::onTabPageStatusChanged);
+  connect(newPage, &TabPage::openDirRequested, this, &MainWindow::onTabPageOpenDirRequested);
+  connect(newPage, &TabPage::sortFilterChanged, this, &MainWindow::onTabPageSortFilterChanged);
+  connect(newPage, &TabPage::backwardRequested, this, &MainWindow::on_actionGoBack_triggered);
+  connect(newPage, &TabPage::forwardRequested, this, &MainWindow::on_actionGoForward_triggered);
 
   ui.tabBar->insertTab(index, newPage->title());
 
@@ -196,6 +269,7 @@ void MainWindow::on_actionGoUp_triggered() {
   TabPage* page = currentPage();
 
   if(page) {
+    ui.filterBar->clear();
     page->up();
     updateUIForCurrentPage();
   }
@@ -205,6 +279,7 @@ void MainWindow::on_actionGoBack_triggered() {
   TabPage* page = currentPage();
 
   if(page) {
+    ui.filterBar->clear();
     page->backward();
     updateUIForCurrentPage();
   }
@@ -214,6 +289,7 @@ void MainWindow::on_actionGoForward_triggered() {
   TabPage* page = currentPage();
 
   if(page) {
+    ui.filterBar->clear();
     page->forward();
     updateUIForCurrentPage();
   }
@@ -238,10 +314,25 @@ void MainWindow::on_actionNewTab_triggered() {
 
 void MainWindow::on_actionNewWin_triggered() {
   FmPath* path = currentPage()->path();
-  Application* app = static_cast<Application*>(qApp);
-  MainWindow* newWin = new MainWindow(path);
-  newWin->resize(app->settings().windowWidth(), app->settings().windowHeight());
-  newWin->show();
+  (new MainWindow(path))->show();
+}
+
+void MainWindow::on_actionNewFolder_triggered() {
+  if(TabPage* tabPage = currentPage()) {
+    FmPath* dirPath = tabPage->folderView()->path();
+
+    if(dirPath)
+      createFileOrFolder(CreateNewFolder, dirPath);
+  }
+}
+
+void MainWindow::on_actionNewBlankFile_triggered() {
+  if(TabPage* tabPage = currentPage()) {
+    FmPath* dirPath = tabPage->folderView()->path();
+
+    if(dirPath)
+      createFileOrFolder(CreateNewTextFile, dirPath);
+  }
 }
 
 void MainWindow::on_actionCloseTab_triggered() {
@@ -285,6 +376,7 @@ void MainWindow::on_actionFolderProperties_triggered() {
 void MainWindow::on_actionShowHidden_triggered(bool checked) {
   TabPage* tabPage = currentPage();
   tabPage->setShowHidden(checked);
+  ui.sidePane->setShowHidden(checked);
   static_cast<Application*>(qApp)->settings().setShowHidden(checked);
 }
 
@@ -324,6 +416,11 @@ void MainWindow::on_actionFolderFirst_triggered(bool checked) {
   currentPage()->setSortFolderFirst(checked);
 }
 
+void MainWindow::on_actionFilter_triggered(bool checked) {
+  ui.filterBar->setVisible(checked);
+  static_cast<Application*>(qApp)->settings().setShowFilter(checked);
+}
+
 void MainWindow::on_actionComputer_triggered() {
   FmPath* path = fm_path_new_for_uri("computer:///");
   chdir(path);
@@ -402,6 +499,32 @@ void MainWindow::onTabBarCloseRequested(int index) {
   closeTab(index);
 }
 
+void MainWindow::onTabBarTabMoved(int from, int to) {
+  // a tab in the tab bar is moved by the user, so we have to move the
+  //  corredponding tab page in the stacked widget to the new position, too.
+  QWidget* page = ui.stackedWidget->widget(from);
+  if(page) {
+	// we're not going to delete the tab page, so here we block signals
+	// to avoid calling the slot onStackedWidgetWidgetRemoved() before
+	// removing the page. Otherwise the page widget will be destroyed.
+    ui.stackedWidget->blockSignals(true);
+    ui.stackedWidget->removeWidget(page);
+    ui.stackedWidget->insertWidget(to, page); // insert the page to the new position
+    ui.stackedWidget->blockSignals(false); // unblock signals
+    ui.stackedWidget->setCurrentWidget(page);
+  }
+}
+
+void MainWindow::onFilterStringChanged(QString str) {
+  if(TabPage* tabPage = currentPage()) {
+    // appy filter only if needed (not if tab is changed)
+    if(str != tabPage->getFilterStr()) {
+      tabPage->setFilterStr(str);
+      tabPage->applyFilter();
+    }
+  }
+}
+
 void MainWindow::closeTab(int index) {
   QWidget* page = ui.stackedWidget->widget(index);
   if(page) {
@@ -415,20 +538,34 @@ void MainWindow::closeTab(int index) {
 void MainWindow::resizeEvent(QResizeEvent *event) {
   QMainWindow::resizeEvent(event);
   Settings& settings = static_cast<Application*>(qApp)->settings();
-  settings.setLastWindowWidth(width());
-  settings.setLastWindowHeight(height());
+  if(settings.rememberWindowSize()) {
+    settings.setLastWindowMaximized(isMaximized());
+
+    if(!isMaximized()) {
+        settings.setLastWindowWidth(width());
+        settings.setLastWindowHeight(height());
+    }
+  }
 }
 
 void MainWindow::closeEvent(QCloseEvent *event)
 {
   QWidget::closeEvent(event);
   Settings& settings = static_cast<Application*>(qApp)->settings();
-  settings.setLastWindowWidth(width());
-  settings.setLastWindowHeight(height());
+  if(settings.rememberWindowSize()) {
+    settings.setLastWindowMaximized(isMaximized());
+
+    if(!isMaximized()) {
+        settings.setLastWindowWidth(width());
+        settings.setLastWindowHeight(height());
+    }
+  }
 }
 
 void MainWindow::onTabBarCurrentChanged(int index) {
   ui.stackedWidget->setCurrentIndex(index);
+  if(TabPage* page = static_cast<TabPage*>(ui.stackedWidget->widget(index)))
+    ui.filterBar->setText(page->getFilterStr());
   updateUIForCurrentPage();
 }
 
@@ -506,6 +643,7 @@ void MainWindow::updateUIForCurrentPage() {
 
     // update side pane
     ui.sidePane->setCurrentPath(tabPage->path());
+    ui.sidePane->setShowHidden(tabPage->showHidden());
 
     // update back/forward/up toolbar buttons
     ui.actionGoUp->setEnabled(tabPage->canUp());
@@ -517,7 +655,6 @@ void MainWindow::updateUIForCurrentPage() {
   }
 }
 
-
 void MainWindow::onStackedWidgetWidgetRemoved(int index) {
   // qDebug("onStackedWidgetWidgetRemoved: %d", index);
   // need to remove associated tab from tabBar
@@ -569,20 +706,16 @@ void MainWindow::onTabPageStatusChanged(int type, QString statusText) {
 
 void MainWindow::onTabPageOpenDirRequested(FmPath* path, int target) {
   switch(target) {
-  case View::OpenInCurrentView:
+  case OpenInCurrentTab:
     chdir(path);
     break;
 
-  case View::OpenInNewTab:
+  case OpenInNewTab:
     addTab(path);
     break;
 
-  case View::OpenInNewWindow: {
-    Application* app = static_cast<Application*>(qApp);
-    MainWindow* newWin = new MainWindow(path);
-    // TODO: apply window size from app->settings
-    newWin->resize(app->settings().windowWidth(), app->settings().windowHeight());
-    newWin->show();
+  case OpenInNewWindow: {
+    (new MainWindow(path))->show();
     break;
   }
   }
@@ -593,26 +726,63 @@ void MainWindow::onTabPageSortFilterChanged() {
 
   if(tabPage == currentPage()) {
     updateViewMenuForCurrentPage();
+    Settings& settings = static_cast<Application*>(qApp)->settings();
+    settings.setSortColumn(static_cast<Fm::FolderModel::ColumnId>(tabPage->sortColumn()));
+    settings.setSortOrder(tabPage->sortOrder());
+    settings.setSortFolderFirst(tabPage->sortFolderFirst());
   }
 }
 
 
 void MainWindow::onSidePaneChdirRequested(int type, FmPath* path) {
-  chdir(path);
+  // FIXME: use enum for type value or change it to button.
+  if(type == 0) // left button (default)
+    chdir(path);
+  else if(type == 1) // middle button
+    addTab(path);
+  else if(type == 2) // new window
+    (new MainWindow(path))->show();
+}
+
+void MainWindow::onSidePaneOpenFolderInNewWindowRequested(FmPath* path) {
+  (new MainWindow(path))->show();
+}
+
+void MainWindow::onSidePaneOpenFolderInNewTabRequested(FmPath* path) {
+  addTab(path);
+}
+
+void MainWindow::onSidePaneOpenFolderInTerminalRequested(FmPath* path) {
+  Application* app = static_cast<Application*>(qApp);
+  app->openFolderInTerminal(path);
+}
+
+void MainWindow::onSidePaneCreateNewFolderRequested(FmPath* path) {
+  createFileOrFolder(CreateNewFolder, path);
+}
+
+void MainWindow::onSidePaneModeChanged(Fm::SidePane::Mode mode) {
+  static_cast<Application*>(qApp)->settings().setSidePaneMode(mode);
+}
+
+void MainWindow::onSplitterMoved(int pos, int index) {
+  Application* app = static_cast<Application*>(qApp);
+  app->settings().setSplitterPos(pos);
 }
 
 void MainWindow::loadBookmarksMenu() {
-  GList* l = fm_bookmarks_get_all(bookmarks);
+  GList* allBookmarks = fm_bookmarks_get_all(bookmarks);
   QAction* before = ui.actionAddToBookmarks;
 
-  for(; l; l = l->next) {
+  for(GList* l = allBookmarks; l; l = l->next) {
     FmBookmarkItem* item = reinterpret_cast<FmBookmarkItem*>(l->data);
-    BookmarkAction* action = new BookmarkAction(item);
-    connect(action, SIGNAL(triggered(bool)), SLOT(onBookmarkActionTriggered()));
+    BookmarkAction* action = new BookmarkAction(item, ui.menu_Bookmarks);
+    connect(action, &QAction::triggered, this, &MainWindow::onBookmarkActionTriggered);
     ui.menu_Bookmarks->insertAction(before, action);
   }
 
   ui.menu_Bookmarks->insertSeparator(before);
+  g_list_free_full(allBookmarks, (GDestroyNotify)fm_bookmark_item_unref);
 }
 
 void MainWindow::onBookmarksChanged(FmBookmarks* bookmarks, MainWindow* pThis) {
@@ -633,9 +803,22 @@ void MainWindow::onBookmarksChanged(FmBookmarks* bookmarks, MainWindow* pThis) {
 void MainWindow::onBookmarkActionTriggered() {
   BookmarkAction* action = static_cast<BookmarkAction*>(sender());
   FmPath* path = action->path();
-
-  if(path)
-    chdir(path);
+  if(path) {
+    Application* app = static_cast<Application*>(qApp);
+    Settings& settings = app->settings();
+    switch(settings.bookmarkOpenMethod()) {
+    case OpenInCurrentTab: /* current tab */
+    default:
+      chdir(path);
+      break;
+    case OpenInNewTab: /* new tab */
+      addTab(path);
+      break;
+    case OpenInNewWindow: /* new window */
+      (new MainWindow(path))->show();
+      break;
+    }
+  }
 }
 
 void MainWindow::on_actionCopy_triggered() {
@@ -664,7 +847,7 @@ void MainWindow::on_actionDelete_triggered() {
 
   bool shiftPressed = (qApp->keyboardModifiers() & Qt::ShiftModifier ? true : false);
   if(settings.useTrash() && !shiftPressed)
-    FileOperation::trashFiles(paths, settings.confirmDelete(), this);
+    FileOperation::trashFiles(paths, settings.confirmTrash(), this);
   else
     FileOperation::deleteFiles(paths, settings.confirmDelete(), this);
 
@@ -729,16 +912,51 @@ void MainWindow::onBackForwardContextMenu(QPoint pos) {
   QAction* selectedAction = menu.exec(btn->mapToGlobal(pos));
   if(selectedAction) {
     int index = menu.actions().indexOf(selectedAction);
+    ui.filterBar->clear();
     page->jumpToHistory(index);
     updateUIForCurrentPage();
   }
 }
 
+void MainWindow::tabContextMenu(const QPoint& pos) {
+  int tabNum = ui.tabBar->count();
+  if(tabNum <= 1) return;
+
+  rightClickIndex = ui.tabBar->tabAt(pos);
+  if(rightClickIndex < 0) return;
+
+  QMenu menu;
+  if(rightClickIndex > 0)
+      menu.addAction(ui.actionCloseLeft);
+  if(rightClickIndex < tabNum - 1) {
+    menu.addAction(ui.actionCloseRight);
+    if(rightClickIndex > 0) {
+      menu.addSeparator();
+      menu.addAction(ui.actionCloseOther);
+    }
+  }
+  menu.exec(ui.tabBar->mapToGlobal(pos));
+}
+
+void MainWindow::closeLeftTabs() {
+  while(rightClickIndex > 0) {
+    closeTab(rightClickIndex - 1);
+    --rightClickIndex;
+  }
+}
+
+void MainWindow::closeRightTabs() {
+  if(rightClickIndex < 0) return;
+  while(rightClickIndex < ui.tabBar->count() - 1)
+    closeTab(rightClickIndex + 1);
+}
+
 void MainWindow::updateFromSettings(Settings& settings) {
   // apply settings
 
   // menu
   ui.actionDelete->setText(settings.useTrash() ? tr("&Move to Trash") : tr("&Delete"));
+  ui.actionDelete->setIcon(settings.useTrash() ? QIcon::fromTheme("user-trash") : QIcon::fromTheme("edit-delete"));
 
   // side pane
   ui.sidePane->setIconSize(QSize(settings.sidePaneIconSize(), settings.sidePaneIconSize()));
@@ -816,6 +1034,25 @@ void MainWindow::on_actionOpenAsRoot_triggered() {
   }
 }
 
+void MainWindow::on_actionFindFiles_triggered() {
+  Application* app = static_cast<Application*>(qApp);
+  FmPathList* selectedPaths = currentPage()->selectedFilePaths();
+  QStringList paths;
+  if(selectedPaths) {
+    for(GList* l = fm_path_list_peek_head_link(selectedPaths); l; l = l->next) {
+      // FIXME: is it ok to use display name here?
+      // This might be broken on filesystems with non-UTF-8 filenames.
+      Fm::Path path(FM_PATH(l->data));
+      paths.append(path.displayName(false));
+    }
+    fm_path_list_unref(selectedPaths);
+  }
+  else {
+    paths.append(currentPage()->pathName());
+  }
+  app->findFiles(paths);
+}
+
 void MainWindow::on_actionOpenTerminal_triggered() {
   TabPage* page = currentPage();
   if(page) {
