@@ -19,23 +19,54 @@
 
 
 #include "tabpage.h"
-#include "filelauncher.h"
-#include "filemenu.h"
-#include "mountoperation.h"
+#include "launcher.h"
+#include <libfm-qt/filemenu.h>
+#include <libfm-qt/mountoperation.h>
 #include <QApplication>
 #include <QCursor>
 #include <QMessageBox>
 #include <QScrollBar>
-#include "proxyfoldermodel.h"
+#include <libfm-qt/proxyfoldermodel.h>
 #include "settings.h"
 #include "application.h"
-#include "cachedfoldermodel.h"
+#include <libfm-qt/cachedfoldermodel.h>
 #include <QTimer>
+#include <QTextStream>
 
 using namespace Fm;
 
 namespace PCManFM {
 
+bool ProxyFilter::filterAcceptsRow(const Fm::ProxyFolderModel* model, FmFileInfo* info) const {
+  if(!model || !info)
+    return true;
+  QString baseName(fm_file_info_get_name(info));
+  if(!virtHiddenList_.isEmpty() && !model->showHidden() && virtHiddenList_.contains(baseName))
+      return false;
+  if(!filterStr_.isEmpty() && !baseName.contains(filterStr_, Qt::CaseInsensitive))
+    return false;
+  return true;
+}
+
+void ProxyFilter::setVirtHidden(FmFolder* folder) {
+  virtHiddenList_ = QStringList(); // reset the list
+  if(!folder) return;
+  if(FmPath* path = fm_folder_get_path(folder)) {
+    char* pathStr = fm_path_to_str(path);
+    if(pathStr) {
+      QString dotHidden = QString(pathStr) + QString("/.hidden");
+      g_free(pathStr);
+      QFile file(dotHidden);
+      if(file.open(QIODevice::ReadOnly | QIODevice::Text)) {
+        QTextStream in(&file);
+        while(!in.atEnd())
+          virtHiddenList_.append(in.readLine());
+        file.close();
+      }
+    }
+  }
+}
+
 TabPage::TabPage(FmPath* path, QWidget* parent):
   QWidget(parent),
   folder_(NULL),
@@ -48,15 +79,20 @@ TabPage::TabPage(FmPath* path, QWidget* parent):
   proxyModel_ = new ProxyFolderModel();
   proxyModel_->setShowHidden(settings.showHidden());
   proxyModel_->setShowThumbnails(settings.showThumbnails());
-  connect(proxyModel_, SIGNAL(sortFilterChanged()), SLOT(onModelSortFilterChanged()));
+  connect(proxyModel_, &ProxyFolderModel::sortFilterChanged, this, &TabPage::onModelSortFilterChanged);
 
   verticalLayout = new QVBoxLayout(this);
   verticalLayout->setContentsMargins(0, 0, 0, 0);
 
   folderView_ = new View(settings.viewMode(), this);
   // newView->setColumnWidth(Fm::FolderModel::ColumnName, 200);
-  connect(folderView_, SIGNAL(openDirRequested(FmPath*, int)), SLOT(onOpenDirRequested(FmPath*, int)));
-  connect(folderView_, SIGNAL(selChanged(int)), SLOT(onSelChanged(int)));
+  connect(folderView_, &View::openDirRequested, this, &TabPage::onOpenDirRequested);
+  connect(folderView_, &View::selChanged, this, &TabPage::onSelChanged);
+  connect(folderView_, &View::clickedBack, this, &TabPage::backwardRequested);
+  connect(folderView_, &View::clickedForward, this, &TabPage::forwardRequested);
+
+  proxyFilter_ = new ProxyFilter();
+  proxyModel_->addFilter(proxyFilter_);
 
   // FIXME: this is very dirty
   folderView_->setModel(proxyModel_);
@@ -66,8 +102,9 @@ TabPage::TabPage(FmPath* path, QWidget* parent):
 }
 
 TabPage::~TabPage() {
-  qDebug("delete TabPage");
   freeFolder();
+  if(proxyFilter_)
+    delete proxyFilter_;
   if(proxyModel_)
     delete proxyModel_;
   if(folderModel_)
@@ -101,7 +138,6 @@ void TabPage::freeFolder() {
     QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
     pThis->overrideCursor_ = true;
   }
-  qDebug("start-loading");
 #if 0
 #if FM_CHECK_VERSION(1, 0, 2) && 0 // disabled
   if(fm_folder_is_incremental(_folder)) {
@@ -121,8 +157,23 @@ void TabPage::freeFolder() {
 #endif
 }
 
-/*static*/ void TabPage::onFolderFinishLoading(FmFolder* _folder, TabPage* pThis) {
+// slot
+void TabPage::restoreScrollPos() {
+  // scroll to recorded position
+  folderView_->childView()->verticalScrollBar()->setValue(browseHistory().currentScrollPos());
 
+  // if the current folder is the parent folder of the last browsed folder,
+  // select the folder item in current view.
+  if(lastFolderPath_.parent() == path()) {
+    QModelIndex index = folderView_->indexFromFolderPath(lastFolderPath_.data());
+    if(index.isValid()) {
+      folderView_->childView()->scrollTo(index, QAbstractItemView::EnsureVisible);
+      folderView_->childView()->setCurrentIndex(index);
+    }
+  }
+}
+
+/*static*/ void TabPage::onFolderFinishLoading(FmFolder* _folder, TabPage* pThis) {
   // FIXME: is this needed?
   FmFileInfo* fi = fm_folder_get_info(_folder);
   if(fi) { // if loading of the folder fails, it's possible that we don't have FmFileInfo.
@@ -132,7 +183,7 @@ void TabPage::freeFolder() {
 
   fm_folder_query_filesystem_info(_folder); // FIXME: is this needed?
 #if 0
-  FmFolderView* fv = pThis->folder_view;
+  FmFolderView* fv = folder_view;
   const FmNavHistoryItem* item;
   GtkScrolledWindow* scroll = GTK_SCROLLED_WINDOW(fv);
 
@@ -156,10 +207,6 @@ void TabPage::freeFolder() {
     g_object_unref(model);
   }
 
-  /* scroll to recorded position */
-  item = fm_nav_history_get_cur(pThis->nav_history);
-  gtk_adjustment_set_value(gtk_scrolled_window_get_vadjustment(scroll), item->scroll_pos);
-
 #endif
 
   // update status text
@@ -171,7 +218,11 @@ void TabPage::freeFolder() {
     QApplication::restoreOverrideCursor(); // remove busy cursor
     pThis->overrideCursor_ = false;
   }
-  qDebug("finish-loading");
+
+  // After finishing loading the folder, the model is updated, but Qt delays the UI update
+  // for performance reasons. Therefore at this point the UI is not up to date.
+  // Of course, the scrollbar ranges are not updated yet. We solve this by installing an Qt timeout handler.
+  QTimer::singleShot(10, pThis, SLOT(restoreScrollPos()));
 }
 
 /*static*/ FmJobErrorAction TabPage::onFolderError(FmFolder* _folder, GError* err, FmJobErrorSeverity severity, TabPage* pThis) {
@@ -187,7 +238,7 @@ void TabPage::freeFolder() {
         // Because the two signals are not correctly paired, we need to
         // remove busy cursor here since "finish-loading" is not emitted.
         QApplication::restoreOverrideCursor(); // remove busy cursor
-	pThis->overrideCursor_ = false;
+        pThis->overrideCursor_ = false;
         return FM_JOB_RETRY;
       }
     }
@@ -288,6 +339,9 @@ void TabPage::chdir(FmPath* newPath, bool addHistory) {
     if(fm_path_equal(newPath, fm_folder_get_path(folder_)))
       return;
 
+    // remember the previous folder path that we have browsed.
+    lastFolderPath_ = fm_folder_get_path(folder_);
+
     if(addHistory) {
       // store current scroll pos in the browse history
       BrowseHistoryItem& item = history_.currentItem();
@@ -311,6 +365,11 @@ void TabPage::chdir(FmPath* newPath, bool addHistory) {
   g_free(disp_name);
 
   folder_ = fm_folder_from_path(newPath);
+  proxyFilter_->setVirtHidden(folder_);
+  if(addHistory) {
+    // add current path to browse history
+    history_.add(path());
+  }
   g_signal_connect(folder_, "start-loading", G_CALLBACK(onFolderStartLoading), this);
   g_signal_connect(folder_, "finish-loading", G_CALLBACK(onFolderFinishLoading), this);
   g_signal_connect(folder_, "error", G_CALLBACK(onFolderError), this);
@@ -322,7 +381,10 @@ void TabPage::chdir(FmPath* newPath, bool addHistory) {
 
   folderModel_ = CachedFolderModel::modelFromFolder(folder_);
   proxyModel_->setSourceModel(folderModel_);
-  proxyModel_->sort(Fm::FolderModel::ColumnFileName);
+  proxyModel_->sort(proxyModel_->sortColumn(), proxyModel_->sortOrder());
+  Settings& settings = static_cast<Application*>(qApp)->settings();
+  proxyModel_->setFolderFirst(settings.sortFolderFirst());
+  proxyModel_->sort(settings.sortColumn(), settings.sortOrder());
 
   if(fm_folder_is_loaded(folder_)) {
     onFolderStartLoading(folder_, this);
@@ -331,12 +393,6 @@ void TabPage::chdir(FmPath* newPath, bool addHistory) {
   }
   else
     onFolderStartLoading(folder_, this);
-
-  if(addHistory) {
-    // add current path to browse history
-    QAbstractItemView* childView = folderView_->childView();
-    history_.add(path());
-  }
 }
 
 void TabPage::selectAll() {
@@ -375,15 +431,13 @@ void TabPage::onSelChanged(int numSel) {
       fm_file_info_list_unref(files);
     }
     else {
-      FmFileInfoList* files;
       goffset sum;
       GList* l;
-      char size_str[128];
       msg = tr("%1 item(s) selected", NULL, numSel).arg(numSel);
       /* don't count if too many files are selected, that isn't lightweight */
       if(numSel < 1000) {
         sum = 0;
-        files = folderView_->selectedFiles();
+        FmFileInfoList* files = folderView_->selectedFiles();
         for(l = fm_file_info_list_peek_head_link(files); l; l = l->next) {
           if(fm_file_info_is_dir(FM_FILE_INFO(l->data))) {
             /* if we got a directory then we cannot tell it's size
@@ -394,6 +448,7 @@ void TabPage::onSelChanged(int numSel) {
           sum += fm_file_info_get_size(FM_FILE_INFO(l->data));
         }
         if(sum >= 0) {
+          char size_str[128];
           fm_file_size_to_str(size_str, sizeof(size_str), sum,
                               fm_config->si_unit);
 	  msg += QString(" (%1)").arg(QString::fromUtf8(size_str));
@@ -411,11 +466,21 @@ void TabPage::onSelChanged(int numSel) {
 
 
 void TabPage::backward() {
+  // remember current scroll position
+  BrowseHistoryItem& item = history_.currentItem();
+  QAbstractItemView* childView = folderView_->childView();
+  item.setScrollPos(childView->verticalScrollBar()->value());
+
   history_.backward();
   chdir(history_.currentPath(), false);
 }
 
 void TabPage::forward() {
+  // remember current scroll position
+  BrowseHistoryItem& item = history_.currentItem();
+  QAbstractItemView* childView = folderView_->childView();
+  item.setScrollPos(childView->verticalScrollBar()->value());
+
   history_.forward();
   chdir(history_.currentPath(), false);
 }
@@ -423,6 +488,11 @@ void TabPage::forward() {
 void TabPage::jumpToHistory(int index)
 {
   if(index >=0 && index < history_.size()) {
+    // remember current scroll position
+    BrowseHistoryItem& item = history_.currentItem();
+    QAbstractItemView* childView = folderView_->childView();
+    item.setScrollPos(childView->verticalScrollBar()->value());
+
     history_.setCurrentIndex(index);
     chdir(history_.currentPath(), false);
   }
@@ -436,8 +506,9 @@ void TabPage::up() {
   FmPath* _path = path();
   if(_path) {
     FmPath* parent = fm_path_get_parent(_path);
-    if(parent)
+    if(parent) {
       chdir(parent, true);
+    }
   }
 }
 
@@ -449,4 +520,19 @@ void TabPage::updateFromSettings(Settings& settings) {
   folderView_->updateFromSettings(settings);
 }
 
-};
+void TabPage::setShowHidden(bool showHidden) {
+  if(!proxyModel_ || showHidden == proxyModel_->showHidden())
+    return;
+  proxyModel_->setShowHidden(showHidden);
+  statusText_[StatusTextNormal] = formatStatusText();
+  Q_EMIT statusChanged(StatusTextNormal, statusText_[StatusTextNormal]);
+}
+
+void TabPage:: applyFilter() {
+  if(!proxyModel_) return;
+  proxyModel_->updateFilters();
+  statusText_[StatusTextNormal] = formatStatusText();
+  Q_EMIT statusChanged(StatusTextNormal, statusText_[StatusTextNormal]);
+}
+
+} // namespace PCManFM
